import re

from wizlib.app import WizApp

from vernum.command import VerNumCommand

FORMAT = re.compile(r"^v?(\d+)\.(\d+)\.(?:(\d+)|beta(\d+)|alpha(\d+))$")


class VerNumApp(WizApp):

    base_command = VerNumCommand
    name = 'vernum'


# def string2val(string):
#     """Convert a string like "1.2.3" to a tuple i.e. (1,2,3)"""
#     if match := FORMAT.match(string):
#         return tuple((-1 if (x is None) else int(x)) for x in match.groups())


# def val2string(major, minor, patch, beta, alpha):
#     """Convert a value list (1,2,3) to a string like "1.2.3" """
#     if alpha >= 0:
#         final = f'alpha{alpha}'
#     elif beta >= 0:
#         final = f'beta{beta}'
#     elif patch >= 0:
#         final = str(patch)
#     return f"{major}.{minor}.{final}"



from argparse import ArgumentParser
from getpass import getpass
from functools import reduce
import sys

from vernum.command import VerNumCommand
from vernum.error import VerNumError
from vernum.scheme import Scheme


class LimitCommand(VerNumCommand):
    """Limit a version number to a minimum"""

    name = 'limit'

    @VerNumCommand.wrap
    def execute(self):
        """Limit a version number to a minimum"""
        limit_str = self.config.get('vernum-limit-min')
        limit = self.scheme.parse(limit_str)
        vernum_str = self.input.text
        vernum = self.scheme.parse(vernum_str)
        if (vernum is None):
            raise VerNumError(f"Invalid version number {vernum_str}")
        if vernum.val < limit.val:
            raise VerNumError(f"Limit violation: {vernum_str} < {limit_str}")
        self.status = f"{vernum_str} >= {limit_str}"

        # current = scheme.parse(self.input)
        # next = getattr(current.increment, self.increment)
        # self.status = f'Version number incremented from {current} to {next}'
        # return str(next)

        # zeroval = (0, 0, 0, -1, -1)
        # verstringlist = self.input.split('\n')
        # vals = [v for s in verstringlist if (v := string2val(s))]
        # maxval = reduce(max, vals, zeroval)
        # self.status = 'Done'
        # return val2string(*maxval)



# from dataclasses import dataclass
import re

from wizlib.class_family import ClassFamily


# def notch(value, up):
#     if type(up) == list:
#         return up[0]
#     else:
#         return value + up


# @dataclass
class Scheme(ClassFamily):
    """Represents a version numbering scheme or system."""

    val: tuple

    def __init__(self, *val):
        """val is a tuple that defines ordering"""
        self.val = val

    def __str__(self):
        """Default string rendering assumes just numbers"""
        if self.val:
            return '.'.join(str(i) for i in self.val)

    @classmethod
    def parse(cls, string):
        """Return an instance by parsing a string"""
        if match := re.match(cls.FORMAT, string):
            groups = match.groups()
            return cls(*tuple(neg4none(x) for x in groups))


def neg4none(value):
    return -1 if (value is None) else int(value)



from dataclasses import dataclass
import re

from vernum.scheme import Scheme


def prelabel(pre: int) -> str:
    return '-alpha' if pre == -2 else '-beta' if pre == -1 else ''


class PatchHyphenAlBeScheme(Scheme):
    """
    Example: 3.4.2-alpha
    """

    name = 'patch-hyphen-al-be'

    FORMAT = re.compile(
        r"^v?(\d+)\.(\d+)\.(\d+)(?:-(alpha|beta))?$")
    INCREMENTS = ['major', 'minor', 'patch', 'beta', 'alpha']

    @classmethod
    def parse(cls, string):
        """Because alpha and beta mean something, must use numbers"""
        if match := re.match(cls.FORMAT, string):
            major, minor, patch, prestr = match.groups()
            pre = -2 if prestr == 'alpha' else -1 if prestr == 'beta' else 0
            return cls(int(major), int(minor), int(patch), pre)

    def __str__(self):
        if self.val:
            major, minor, patch, pre = self.val
            result = f"{str(major)}.{str(minor)}.{str(patch)}"
            result += prelabel(pre)
            return result

    def increment(self, up: str):
        major, minor, patch, pre = self.val
        if up == 'alpha':
            return
        elif up == 'beta' and pre == -2:
            pre = -1
        elif up == 'patch':
            patch += 1
            pre = -2
        elif up == 'patch-zero':
            if pre == 0:
                patch += 1
            pre = 0
        elif up == 'minor':
            minor += 1
            patch = 0
            pre = -2
        elif up == 'minor-zero':
            if patch > 0 or pre == 0:
                minor += 1
            patch = pre = 0
        elif up == 'major':
            major += 1
            minor = patch = 0
            pre = -2
        elif up == 'major-zero':
            if minor > 0 or patch > 0 or pre == 0:
                major += 1
            minor = patch = pre = 0
        return PatchHyphenAlBeScheme(major, minor, patch, pre)



from vernum.scheme import Scheme


class MinorScheme(Scheme):
    """Two part version numbers always with major and minor"""

    name = 'minor'

    FORMAT = r"^v?(\d+)\.(\d+)$"
    INCREMENTS = ['major', 'minor']

    @classmethod
    def zero(cls):
        return cls(0, 0)

    def increment(self, up: str):
        major, minor = self.val
        if up == 'minor':
            minor += 1
        elif up == 'major':
            major += 1
            minor = 0
        return MinorScheme(major, minor)
